#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
EndPoints

Base classes to implement message endpoints (EIP2003)
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.
from functools import partial

__all__ = ['Endpoint', 'Replier', 'Requestor', 'Replay']

# Project requirements
from sleipnir.core.decorators import cached
from sleipnir.core.utils import maybe_list, uuid

# local submodule requirements
from ..connections import ConnectionFactory
from ..amqp.channels import Producer, Consumer, ChannelEnd

RELAY = 'relay'
REPLY_TO = 'reply_to'
MESSAGE_ID = 'message_id'
CORRELATION_ID = 'correlation_id'


class EndPoint(object):

    # default consumer options to be used
    # when a new Consumer is created
    consumer = Consumer
    consumer_opt = {}

    # default consumer options to be used
    # when a new Producer is created
    producer = Producer
    producer_opt = {}

    # invalid exchange options
    invalid = None
    invalid_opt = {}

    def __init__(self, request, reply=None, error=None, connection=None):
        #setup connection
        factory = ConnectionFactory.get_instance()
        self._connection = conn = factory.create(connection)

        #create routes
        factory = ChannelEnd.create
        error = factory(error, self.invalid, self.invalid_opt, conn)
        consumer = factory(request, self.consumer, self.consumer_opt, conn)
        producer = factory(reply, self.producer, self.producer_opt, conn)
        # store routes
        self._error, self._consumer, self._producer = error, consumer, producer

    def reply(self, response, message, reply=None):
        reply = reply or self._producer or self._error
        reply and reply.publish((response, message,))

    def run(self, timeout=None):
        self._consumer.consume(timeout) if self._consumer else None

    @property
    @cached
    def nodes(self):
        class Ends(object):
            def __init__(self, producer, consumer, error):
                self.consumer = consumer
                self.producer = producer
                self.error = error
        return Ends(self._producer, self._consumer, self._error)


class _Wrapper(object):
    def __init__(self, callback=None):
        self._callback = callback

    def __getattr__(self, value):
        return partial(self.__invoke, value)

    def __invoke(self, method, *args, **kwargs):
        method = getattr(self._callback, method, None)
        return method(*args, **kwargs) if method else args[0]


class Replier(EndPoint):

    def __init__(self, request, error=None, connection=None, callback=None):
        super(Replier, self).__init__(request, None, error, connection)
        self._repliers = {}
        self._callback = _Wrapper(callback)
        self._consumer.append_callbacks(self.process)

    def _get_replier(self, address):
        if address not in self._repliers:
            exchange = address if '@' in address else '@' + address
            self._repliers[address] = ChannelEnd.create(
                exchange,
                self.producer,
                self.producer_opt,
                self._connection)
        # get valid replier
        return self._repliers[address]

    def process(self, request, message):
        # fetch data
        payload, props = message.payload, message.properties
        reply_to, message_id = props[REPLY_TO], props[MESSAGE_ID]
        # send to replier
        payload, props = self._callback.process(payload, props)
        if reply_to is not None:
            # publish in replier
            props.setdefault(MESSAGE_ID, uuid())
            props.setdefault(CORRELATION_ID, message_id)
            replier = self._get_replier(reply_to)
            # prepare message
            message = replier.prepare(payload, **props)
            message = self._callback.prepared(message)
            # publish reply
            replier.publish(message)
            self._callback.published(message)


class Requestor(EndPoint):
    def __init__(self, request, response,
                 error=None, connection=None, callback=None):
        super(Requestor, self).__init__(request, response, error, connection)
        self._callback = _Wrapper(callback)
        self._consumer.append_callbacks(self._callback.response)

    def request(self, request=None, headers={}):
        requests = self._callback.request(request, headers)
        messages = requests if hasattr(requests, 'next') else [requests]
        # publish request
        for request, headers in messages:
            message = self._producer.prepare(request, **headers)
            message = self._callback.prepared(message)
            self._producer.publish(message)
            self._callback.published(request, message)

    def run(self, timeout=None):
        self._producer.pending += [
            self.request,
            partial(super(Requestor, self).run, timeout),
        ]


class Replay(EndPoint):
    def __init__(self, request, reply=None, error=None, connection=None):
        self.consumer_opts[RELAY] = True
        super(Replay, self).__init__(request, None, error, connection)

    def process(self, request, message):
        self.reply(request, message)


class Manager(object):
    def __init__(self, endpoints=None):
        self._processes = {}
        self._producers = {}
        self._consumers = {}
        # store active processes here
        for end in maybe_list(endpoints):
            if hasattr(end, 'consume'):
                callback = partial(self.process, end)
                end.append_callbacks(callback)
                self._consumers[end.queues[0]] = end
            if hasattr(end, 'publish'):
                self._producers[end.exchange] = end

    def __getattr__(self, value):
        return getattr(self._processes, value)

    def when_process_completed(self, process_key):
        del self._processes[process_key]

    def run(self, timeout=None):
        [cons.consume(timeout) for cons in self._consumers.itervalues()]

    def process(self, end, request, message):
        raise NotImplementedError



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Routes

Represents a generic gateway to AMQP implementation
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = ['Exchange', 'Queue', 'Route']

# Project requirements
from sleipnir.core.decorators import cached

# local submodule requirements
from ..enums import RouteEndType, RouteEndPosition, RouteState, RouteType
from ..connections import ConnectionFactory
from ..parser import binding_url, URI


class RouteAbstract(object):

    ignore = []

    def __init__(self, params={}, convert={}):
        self._params = self._parse_params(params, convert)

    def __hash__(self):
        return hash(str(self))

    def __iter__(self):
        yield 'end_type', self._type
        for key, value in self._params.iteritems():
            yield key, value

    def __getattr__(self, value):
        return self._params[value]

    def _parse_params(self, params, convert):
        # convert using convert functions
        pseudo_cast = lambda v: v
        convert = convert.get(self._type, {})
        new_params = params if not convert else {}
        for key, native in convert.iteritems():
            if key in params and key not in self.ignore:
                value = (native[1] or pseudo_cast)(params[key])
                if native[2]:
                    new_params.setdefault(native[2], {})
                    new_params[native[2]][native[0]] = value
                else:
                    new_params[native[0]] = value
        return new_params

    @property
    def params(self):
        return self._params


class RouteEnd(RouteAbstract):

    ignore = ['origin', 'destination']

    def __init__(self, params={}, convert={}):
        self._name = params.get('origin', '')
        if 'origin' in self.ignore:
            self._name = params.get('destination', '')
        super(RouteEnd, self).__init__(params, convert)

    def __str__(self):
        return "%s|%s" % (self._type.enumname[0], self._name)

    @property
    def end_type(self):
        return self._type

    @property
    def name(self):
        return self._name

    @staticmethod
    def get_type(params):
        rtype = RouteEndType.QUEUE
        if 'exchage' in params.get('type', ''):
            rtype = RouteEndType.EXCHANGE
        return rtype


class Exchange(RouteEnd):

    _type = RouteEndType.EXCHANGE

    def __init__(self, pos=RouteEndPosition.ORIGIN, args={}, convert={}):
        self.ignore = self.ignore[:]
        self.ignore.remove(pos.enumname.lower())
        super(Exchange, self).__init__(args, convert)

    def __str__(self):
        return "E|%s@%s" % (self._name, self.type)


class Queue(RouteEnd):

    _type = RouteEndType.QUEUE

    def __init__(self, pos=RouteEndPosition.DESTINATION, args={}, convert={}):
        self.ignore = self.ignore[:]
        self.ignore.remove(pos.enumname.lower())
        super(Queue, self).__init__(args, convert)


class Route(RouteAbstract):

    def __init__(self, route, state=RouteState.UNKNOWN):
        # route URI
        self._url = URI(route, binding_url)
        # Route Ends
        self._origin = self._destination = None
        # Rotute Params
        self._params = None
        # Convertion attrs
        self._convert = {}
        # Route state
        self._state = state
        # Route type
        self._type = Route.get_type(self.type)

    def __iter__(self):
        for key, value in self._url:
            if key in ("hosts", "user", "password",):
                continue
            yield key, value

    def __str__(self):
        return "R|%s@%s" % (self._url.origin, self._url.destination)

    def __repr__(self):
        return str(self._url)

    def __call__(self, convert):
        if convert != self._convert:
            self._origin = self._destination = None
            self._convert = convert
        return self

    @cached
    def params(self, convert={}):
        return self._parse_params(dict(iter(self._url)), convert)

    @cached
    def _end_params(self):
        return dict(iter(self._url))

    @property
    def name(self):
        return self._url.name

    @property
    def type(self):
        return self._url.type

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def route_type(self):
        return self._type

    @property
    def origin(self):
        if not self._origin:
            params = self._end_params()
            pos = RouteEndPosition.ORIGIN
            self._origin = Exchange(pos, params, self._convert)
        return self._origin

    @property
    def destination(self):
        if not self._destination:
            params = self._end_params()
            pos = RouteEndPosition.DESTINATION
            if RouteEnd.get_type(params) == RouteEndType.QUEUE:
                self._destination = Queue(pos, params, self._convert)
            else:
                self._destination = Exchange(pos, params, self._convert)
        return self._destination

    @property
    def url(self):
        return str(self._url)

    @staticmethod
    def get_type(str_type):
        rtype = RouteType.QUEUE
        if str_type.startswith('exchage'):
            rtype = RouteType.EXCHANGE
        return rtype



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
enums

Common enumerates used by transport modules
"""

from __future__ import absolute_import

__author__  = "Carlos Martin"
__license__ = "BSD, See LICENSE file for details"

# Import here any required modules.
import sleipnir.core

__all__ = [
    'ChannelTarget',
    'ChannelTargetState',
    'ConnectionState',
    'ConnectionType',
    'DeliveryMode',
    'RouteEndPosition',
    'RouteEndType',
    'RouteState',
    'RouteType',
    ]


class ChannelTarget(enum):
    ENTITY,    \
    ROUTE = xrange(0, 2)


class ChannelTargetState(enum):
    DECLARED,  \
    RESUMED,   \
    SUSPENDED, \
    DESTROYED = xrange(0, 4)


class ConnectionState(enum):
    UNKNOWN,   \
    CLOSED,    \
    STARTING,  \
    STARTED,   \
    CLOSING = xrange(0, 5)


class ConnectionType(enum):
    AUTO_ASYNC,     \
    AUTO_BLOCKING,  \
    AMQP_ASYNC,     \
    AMQP_BLOCKING,  \
    AMQP_MOCK,      \
    AMQP_TORNADO,   \
    HTTP_BLOCKING,  \
    HTTP_ASYNC = xrange(0, 8)


class DeliveryMode(enum):
    TRANSIENT, PERSISTENT = xrange(1, 3)


class RouteEndPosition(enum):
    ORIGIN,    \
    DESTINATION = xrange(0, 2)


class RouteEndType(enum):
    EXCHANGE,  \
    QUEUE = xrange(0, 2)


class RouteType(enum):
    EXCHANGE,  \
    QUEUE = xrange(2, 4)


class RouteState(enum):
    BIND,                  \
    UNBIND,                \
    DESTINATION_DECLARED,  \
    ORIGIN_DECLARED,       \
    UNKNOWN = xrange(0, 5)



#!/usr/bin/env python
# -*- mode:python; tab-width: 2; coding: utf-8 -*-

"""
Routers

Base classes to implement message routing patters (EIP2003)
"""

from __future__ import absolute_import

__author__  = "Carlos Martin <cmartin@liberalia.net>"
__license__ = "See LICENSE file for details"

# Import here any required modules.

__all__ = ['Routers']

# Project requirements

REPLY_TO = "reply_to"

# local submodule requirements
from .endpoints import EndPoint
from ..connections import ConnectionFactory
from ..amqp.channels import ChannelEnd, Consumer, Producer


class Relay(object):

    # replay factory option
    producer = Producer
    producer_opt = {}

    def __init__(self, callback):
        # init connection stuff
        self._callback = callback
        self._routing = {}

    def __getattr__(self, value):
        return getattr(self._routing, value)

    def replay(self, message, properties, receivers=None):
        rcvs = receivers if receivers is not None else self._routing.iterkeys()
        for recipient in rcvs:
            producer = self._routing[recipient][1]
            producer.publish(producer.prepare(message, **properties))

    def add_route(self, route, contents):
        # remove replayer if a empty message is send
        if route in self._routing and len(contents) == 0:
            del self._routing[route]
            return
        # otherwise create producer and and to routes
        publisher = self._callback(
            route,
            self.producer,
            self.producer_opt)
        self._routing[route] = (contents, publisher,)


class EndPointRelay(EndPoint):

    # control queue options
    control = Consumer
    control_opt = {}

    def __init__(self, request=None, control=None,
                 error=None, connection=None, factory=ChannelEnd.create):

        # init consumer and setup callbacks
        super(EndPointRelay, self).__init__(request, None, error, connection)
        # set up relay
        self._relay = Relay(lambda x, y, z: factory(x, y, z, connection))
        # setup control and relay compose
        self._control = ChannelEnd.create(
            control,
            self.control,
            self.control_opt,
            connection)
        self._control  and self._control.append_callbacks(self.process_control)
        self._consumer and self._consumer.append_callbacks(self.process)

    def __getattr__(self, value):
        return getattr(self._relay, value)

    def run(self, timeout=None):
        self._control.consume(timeout)
        super(EndPointRelay, self).run(timeout)

    def process(self, request, response, receivers=None):
        self._relay.replay(request, response, receivers)

    def process_control(self, request, message):
        route = message.properties[REPLY_TO]
        if route is not None:
            self._relay.add_route(route, message.payload)
